from typing import List, Optional, TypedDict
from ..file_utils import convert_file
from ..widget_base import Input, MultipleHandler


class CardOption(TypedDict):
    title: Optional[str]
    subtitle: Optional[str]
    image: Optional[str]
    description: Optional[str]
    topLeftExtra: Optional[str]
    topRightExtra: Optional[str]


class CardsInput(Input):
    type = "cards-input"
    empty_value = None
    multiple: bool = False
    multiple_handler: MultipleHandler

    def __init__(self, key: str, label: str, options: List[CardOption], **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, options=options, **kwargs))

    def set_props(self, props):
        self.label = props.get("label", "Label")
        self.options = [
            {**opt, "image": convert_file(opt.get("image"))}
            for opt in props.get("options", [])
        ]
        self.searchable = props.get("searchable", False)
        self.required = props.get("required", True)
        self.hint = props.get("hint", None)
        self.columns = props.get("columns", 2)
        self.full_width = props.get("full_width", False)
        self.layout = props.get("layout", "list")
        self.disabled = props.get("disabled", False)
        self.multiple = props.get("multiple", False)
        self.empty_value = [] if self.multiple else None
        self.value = props.get("initial_value", self.empty_value)
        self.min = props.get("min", None)
        self.max = props.get("max", None)
        self.multiple_handler = MultipleHandler(
            self.multiple, self.min, self.max, self.required
        )

    def render(self, ctx: dict):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "hint": self.hint,
            "options": self.options,
            "multiple": self.multiple,
            "searchable": self.searchable,
            "required": self.required,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "layout": self.layout,
            "disabled": self.disabled,
            "errors": self.errors,
            "min": self.min,
            "max": self.max,
            "value": self.serialize_value(),
        }

    def validate(self) -> List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) -> List:
        return self.multiple_handler.value_to_list(self.value)

    def parse_value(self, value):
        return self.multiple_handler.value_to_list_or_value(value)



from .TimeInput import TimeInput
from .TextareaInput import TextareaInput
from .PandasRowSelectionInput import PandasRowSelectionInput
from .LatexOutput import LatexOutput
from .CurrencyInput import CurrencyInput
from .FileInput import FileInput
from .CardsInput import CardsInput
from .CpfInput import CpfInput
from .ImageOutput import ImageOutput
from .EmailInput import EmailInput
from .MarkdownOutput import MarkdownOutput
from .TagInput import TagInput
from .DateInput import DateInput
from .NumberInput import NumberInput
from .ProgressOutput import ProgressOutput
from .RatingInput import RatingInput
from .CustomInput import CustomInput
from .ChecklistInput import ChecklistInput
from .NumberSliderInput import NumberSliderInput
from .CnpjInput import CnpjInput
from .CameraInput import CameraInput
from .NpsInput import NpsInput
from .PhoneInput import PhoneInput
from .ToggleInput import ToggleInput
from .PasswordInput import PasswordInput
from .RichTextInput import RichTextInput
from .ListInput import ListInput
from .ImageInput import ImageInput
from .MultipleChoiceInput import MultipleChoiceInput
from .IframeOutput import IframeOutput
from .TextOutput import TextOutput
from .CodeInput import CodeInput
from .VideoInput import VideoInput
from .DropdownInput import DropdownInput
from .FileOutput import FileOutput
from .PlotlyOutput import PlotlyOutput
from .CheckboxInput import CheckboxInput
from .HtmlOutput import HtmlOutput
from .PandasOutput import PandasOutput
from .TextInput import TextInput
from .LinkOutput import LinkOutput



from ..linter import LinterRule, LinterIssue
from typing import Set, List, Dict, Tuple
import ast
from pathlib import Path
from ...repositories.project.project import ProjectRepository
from ...utils.code import function_called_args


class UnsetDataFound(LinterIssue):
    def __init__(self, key: str, path: Path, lineno: int):
        self.label = f"There is a call to get_data('{key}') in {path}:{lineno}, but no corresponding set_data call was found."
        self.fixes = []


class UnsetGetData(LinterRule):
    label: str = "Possibly undefined get data"
    type: str = "info"

    def find_issues(self) -> List[LinterIssue]:
        project = ProjectRepository.load()
        data_gets: Dict[str, Set[Tuple[Path, int]]] = {}
        data_sets: Dict[str, Set[Tuple[Path, int]]] = {}
        for python_file in project.project_files:
            try:
                get_data_calls = function_called_args(
                    python_file.read_text(encoding="utf-8"),
                    ["abstra", "workflows"],
                    "get_data",
                )
                set_data_calls = function_called_args(
                    python_file.read_text(encoding="utf-8"),
                    ["abstra", "workflows"],
                    "set_data",
                )

                if get_data_calls is not None:
                    for function_call in get_data_calls:
                        if len(function_call) > 0:
                            key_arg = function_call[0]
                            if isinstance(key_arg, ast.Constant) and isinstance(
                                key_arg.value, str
                            ):
                                data_gets.setdefault(key_arg.value, set()).add(
                                    (python_file, key_arg.lineno)
                                )
                if set_data_calls is not None:
                    for function_call in set_data_calls:
                        if len(function_call) > 0:
                            key_arg = function_call[0]
                            if isinstance(key_arg, ast.Constant) and isinstance(
                                key_arg.value, str
                            ):
                                data_sets.setdefault(key_arg.value, set()).add(
                                    (python_file, key_arg.lineno)
                                )
            except SyntaxError:
                continue
            except FileNotFoundError:
                continue

        issues = []
        for data_get_key, data_get_path in data_gets.items():
            if data_get_key not in data_sets:
                for path, lineno in data_get_path:
                    issues.append(UnsetDataFound(data_get_key, path, lineno))

        return issues



from typing import List, Union
from ..widget_base import (
    Input,
    LabelValueOption,
    MultipleHandler,
    OptionsHandler,
    AbstraOption,
)


class DropdownInput(Input):
    type = "dropdown-input"
    empty_value = None
    multiple: bool = False
    multiple_handler: MultipleHandler
    options_handler: OptionsHandler

    def __init__(self, key: str, label: str, options: List[AbstraOption], **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, options=options, **kwargs))

    def set_props(self, props):
        self.label = props.get("label", "Label")
        self.options = props.get("options", [])
        self.required = props.get("required", True)
        self.hint = props.get("hint", None)
        self.placeholder = props.get("placeholder", "")
        self.full_width = props.get("full_width", False)
        self.disabled = props.get("disabled", False)
        self.multiple = props.get("multiple", False)
        self.min = props.get("min", None)
        self.max = props.get("max", None)
        self.empty_value = [] if self.multiple else None
        self.value = props.get("initial_value", self.empty_value)
        self.multiple_handler = MultipleHandler(
            self.multiple, self.min, self.max, self.required
        )
        self.options_handler = OptionsHandler(self.options)

    def render(self, ctx: dict):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "options": self.options_handler.serialized_options(),
            "hint": self.hint,
            "multiple": self.multiple,
            "placeholder": self.placeholder,
            "value": self.serialize_value(),
            "required": self.required,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
            "min": self.min,
            "max": self.max,
        }

    def validate(self) -> List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) -> List:
        list_values = self.multiple_handler.value_to_list(self.value)
        return self.options_handler.uuids_from_values(list_values)

    def parse_value(self, value: Union[List[str], None]):
        list_values = self.options_handler.values_from_uuids(value)
        return self.multiple_handler.value_to_list_or_value(list_values)



from io import BufferedReader, TextIOWrapper
from pathlib import Path
from typing import List, Union
from ..apis import upload_file
from ..response_types import FileResponse
from ..widget_base import Input, MultipleHandler


class FileInput(Input):
    type = "file-input"
    multiple: bool = False
    multiple_handler: MultipleHandler

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get("label", "Label")
        self.required = props.get("required", True)
        self.hint = props.get("hint", None)
        self.full_width = props.get("full_width", False)
        self.disabled = props.get("disabled", False)
        self.max_file_size = props.get("max_file_size", None)
        self.multiple = props.get("multiple", False)
        self.value = props.get("initial_value", self.empty_value)
        self.multiple_handler = MultipleHandler(self.multiple, self.required)

    def render(self, ctx: dict):
        return {
            "type": self.type,
            "key": self.key,
            "hint": self.hint,
            "label": self.label,
            "value": self.serialize_value(),
            "required": self.required,
            "multiple": self.multiple,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "maxFileSize": self.max_file_size,
            "errors": self.errors,
        }

    @staticmethod
    def __get_file_uri(
        value: Union[FileResponse, str, BufferedReader, TextIOWrapper]
    ) -> str:
        if isinstance(value, str):
            if Path(value).is_file():
                return upload_file(open(value))
            else:
                return value
        if isinstance(value, FileResponse):
            return value.url
        if isinstance(value, BufferedReader):
            return upload_file(value)
        if isinstance(value, TextIOWrapper):
            return upload_file(value)
        return ""

    def serialize_value(self) -> List[str]:
        values_list = self.multiple_handler.value_to_list(self.value)
        return [FileInput.__get_file_uri(item) for item in values_list]

    def parse_value(
        self, value: List[str]
    ) -> Union[FileResponse, List[FileResponse], None]:
        file_responses = [FileResponse(item) for item in value]
        return self.multiple_handler.value_to_list_or_value(file_responses)



from unittest import TestCase
from .unset_get_data import UnsetGetData
from tests.fixtures import init_dir, clear_dir
from ...repositories.project.project import ProjectRepository, ScriptStage


class UnsetGetDataTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_unset_get_data_valid_default(self):
        rule = UnsetGetData()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_unset_get_data_invalid(self):
        rule = UnsetGetData()
        project = ProjectRepository.load()
        code_path = self.root / "test.py"
        code_path.write_text(
            "\n".join(["from abstra.workflows import get_data", "get_data('test')"])
        )
        script = ScriptStage(
            id="test",
            file=str(code_path),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(script)
        ProjectRepository.save(project)
        self.assertEqual(len(rule.find_issues()), 1)

    def test_unset_get_data_invalid_with_relative_import(self):
        rule = UnsetGetData()
        project = ProjectRepository.load()
        code_path = self.root / "test.py"
        code_path.write_text(
            "\n".join(["import abstra.workflows as aw", "aw.get_data('test')"])
        )
        script = ScriptStage(
            id="test",
            file=str(code_path),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(script)
        ProjectRepository.save(project)
        self.assertEqual(len(rule.find_issues()), 1)

    def test_unset_get_data_valid_with_use_in_different_files_with_relative_import(
        self,
    ):
        rule = UnsetGetData()
        project = ProjectRepository.load()
        setter_file = self.root / "setter.py"
        setter_file.write_text(
            "\n".join(
                [
                    "import abstra.workflows as aw",
                    "aw.set_data('test', 1)",
                ]
            )
        )
        setter_script = ScriptStage(
            id="test",
            file=str(setter_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(setter_script)

        getter_file = self.root / "getter.py"
        getter_file.write_text(
            "\n".join(
                [
                    "from abstra.workflows import get_data",
                    "get_data('test')",
                ]
            )
        )
        getter_script = ScriptStage(
            id="test",
            file=str(getter_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(getter_script)

        ProjectRepository.save(project)
        self.assertEqual(len(rule.find_issues()), 0)

    def test_unset_get_data_check_the_right_argument(
        self,
    ):
        rule = UnsetGetData()
        project = ProjectRepository.load()
        setter_file = self.root / "setter.py"
        setter_file.write_text(
            "\n".join(
                [
                    "import abstra.workflows as aw",
                    "aw.set_data('test', 'bar')",
                ]
            )
        )
        setter_script = ScriptStage(
            id="test",
            file=str(setter_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(setter_script)

        getter_file = self.root / "getter.py"
        getter_file.write_text(
            "\n".join(
                [
                    "from abstra.workflows import get_data",
                    "get_data('bar')",
                ]
            )
        )
        getter_script = ScriptStage(
            id="test",
            file=str(getter_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(getter_script)

        ProjectRepository.save(project)
        self.assertEqual(len(rule.find_issues()), 1)



import flask
from ...repositories import requirements_repository
from ...usage import usage


def get_editor_bp():
    bp = flask.Blueprint("editor_requirements", __name__)

    # 1s pooling in this route
    @bp.get("/")
    def _get_requirements():
        return requirements_repository.load().to_dict()

    @bp.post("/")
    @usage
    def _create_requirement():
        data = flask.request.json

        if not data:
            flask.abort(400)

        name = data["name"]
        version = data.get("version")

        requirements = requirements_repository.load()
        requirements.add(name, version)
        requirements_repository.save(requirements)
        return requirements.to_dict()

    @bp.delete("/<name>")
    @usage
    def _delete_requirement(name: str):
        requirements = requirements_repository.load()
        requirements.delete(name)
        requirements_repository.save(requirements)
        return requirements.to_dict()

    # 1s pooling in this route
    @bp.get("/recommendations")
    def _get_requirements_recommendation():
        return [r.to_dict() for r in requirements_repository.get_recommendation()]

    return bp



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

import requests
from pydantic.dataclasses import dataclass

from ..utils.datetime import from_utc_iso_string, to_utc_iso_string
from ..utils.environment import SIDECAR_HEADERS, SIDECAR_URL
from ..utils.dot_abstra import LOCAL_LOGS_FOLDER


class LogEntryFactory:
    @staticmethod
    def from_dto(dto: Dict) -> "LogEntry":
        if dto.get("event") == "stdout" or dto.get("event") == "stderr":
            return StdioLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                event=dto["event"],
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        elif dto.get("event") == "form-message":
            return FormEventLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        elif dto.get("event") == "unhandled-exception":
            return UnhandledExceptionLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        else:
            raise Exception("Invalid log entry type")


@dataclass
class StdioLogEntry:
    execution_id: str
    created_at: datetime
    event: Literal["stderr", "stdout"]
    payload: Dict[Literal["text"], str]
    sequence: int

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


@dataclass
class FormEventLogEntry:
    execution_id: str
    created_at: datetime
    sequence: int
    payload: Dict[Literal["messageType", "messagePayload"], Union[str, Dict]]
    event: Literal["form-message"] = "form-message"

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


@dataclass
class UnhandledExceptionLogEntry:
    execution_id: str
    created_at: datetime
    payload: Dict[Literal["text"], str]
    sequence: int
    event: Literal["unhandled-exception"] = "unhandled-exception"

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


LogEntry = Union[StdioLogEntry, FormEventLogEntry, UnhandledExceptionLogEntry]
LogEvent = Literal["stdout", "stderr", "form-message", "unhandled-exception"]


class ExecutionLogsRepository(ABC):
    @abstractmethod
    def save(self, log_entry: LogEntry) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        raise NotImplementedError()


class LocalExecutionLogsRepository(ExecutionLogsRepository):
    _logs: Dict[str, List[LogEntry]]

    def __init__(self):
        self._logs = {}

    def save(self, log_entry: LogEntry) -> None:
        execution_id = log_entry.execution_id

        if not execution_id:
            return

        log_file = Path(LOCAL_LOGS_FOLDER) / f"{execution_id}.log"

        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True)

        with open(log_file, "a", encoding="utf-8") as f:
            f.write(json.dumps(log_entry.to_dto()) + "\n")

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        log_file = Path(LOCAL_LOGS_FOLDER) / f"{execution_id}.log"

        if not log_file.exists():
            return []

        try:
            with open(log_file, "r", encoding="utf-8") as f:
                logs: List[LogEntry] = []
                for line in f.readlines():
                    dto = json.loads(line)
                    if event and dto["event"] != event:
                        continue
                    logs.append(LogEntryFactory.from_dto(dto))
                return logs
        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error reading log file {log_file}: {e}")
            return []


class RemoteExecutionLogsRepository(ExecutionLogsRepository):
    def __init__(self, url: str, headers: Dict[str, str]):
        self.url = url
        self.headers = headers

    def save(self, log_entry: LogEntry) -> None:
        res = requests.post(
            f"{self.url}/executions/{log_entry.execution_id}/logs",
            json=log_entry.to_dto(),
            headers=self.headers,
        )

        res.raise_for_status()

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        response = requests.get(
            f"{self.url}/executions/{execution_id}/logs",
            params={"event": event} if event else None,
            headers=self.headers,
        )

        response.raise_for_status()

        return [LogEntryFactory.from_dto(log) for log in response.json()]


def execution_logs_repository_factory() -> ExecutionLogsRepository:
    if SIDECAR_URL:
        return RemoteExecutionLogsRepository(url=SIDECAR_URL, headers=SIDECAR_HEADERS)
    else:
        return LocalExecutionLogsRepository()
